import requests


class NSession(requests.Session):
    def __init__(self):
        super().__init__()
        self._session = requests.Session()
        self.default_timeout = 60

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self._session.close()

    def close(self):
        self._session.close()

    def get(self, url, **kwargs):

        if kwargs.get("timeout", None) is None:
            kwargs.update({"timeout": self.default_timeout})

        return self._session.get(url=url, **kwargs)

    def post(self, url, data=None, json: {}=None, **kwargs):
        """
        :param url:
        :param data:
        :param json: Dictionary object. Send your object using my_obj.__dict__
        :param kwargs:
        :return:
        """

        if kwargs.get("timeout", None) is None:
            kwargs.update({"timeout": self.default_timeout})

        return self._session.post(url=url, data=data, json=json, **kwargs)

    def delete(self, url, **kwargs):

        if kwargs.get("timeout", None) is None:
            kwargs.update({"timeout": self.default_timeout})

        return self._session.delete(url=url, **kwargs)

    def put(self, url, data=None, **kwargs):

        if kwargs.get("timeout", None) is None:
            kwargs.update({"timeout": self.default_timeout})

        return self._session.put(url=url, data=data, **kwargs)

    def add_header(self, name, value):
        self._session.headers.update({name: value})



import json
from datetime import datetime
from typing import List, Optional
from urllib.parse import quote

import time
from isodate import duration_isoformat
from requests import Response

from gateways.apis.api_route_base import ApiRoot
from gateways.common.cs18_api_classes import BlueprintRepositoryDetails, AccessLink, \
    AddBlueprintUsingTokenRepositoryRequest, AddBlueprintGithubRepositoryRequest, \
    AddBlueprintBitbucketRepositoryRequest, BitbucketBlueprintRepositoryDetails
from gateways.common.cs18_api_converters import Converters
from gateways.common.cs18_api_errors import UnauthorizedException, SandboxNotFound, \
    SandboxEndingFailed
from gateways.common.cs18_api_requests import (
    CreateInvitationsRequest,
    UpdateSpaceRequest,
    ProductionEnvironment, DebuggingServiceValue, SpaceCloudAccountInfraSettingsRequest,
    AddK8SComputeServiceToSpaceRequest, UpdateK8SComputeServiceInSpaceRequest, AwsCloudProviderSettingsModel,
    NewUserSurveyRequest)
from gateways.common.cs18_api_responses import (AccountStatusResponse, BlueprintFileResponse, BlueprintResponse,
                                                CatalogForGetAllResponse, CatalogForGetResponse, CloudAccountResponse,
                                                CreateProductionResponse, CreateSandboxResponse, GetSpaceResponse,
                                                GetSpacesResponse, ProductionBlueResponse, ProductionGreenResponse,
                                                ProductionResponseLean, RepositoryResponse, RoleListItemResponse,
                                                SandboxResponse, SandboxResponseLean, TokenResponse,
                                                UserInvitationResponse, UserPermittedToSpaceResponse,
                                                UserForAllUsersResponse, CloudAccountInSpaceResponse, RegionResponse,
                                                VerifyCloudProviderResponse, AccountRepositoryResponse,
                                                BlueprintValidationResponse, AccountRepositoryDetailsResponse,
                                                DirectAwsSshResponse, DirectAzureSshResponse, DirectRdpResponse,
                                                ParameterStoreItem, RepositoryProviderResponse,
                                                RepositoryRelationshipResponse, AccountExtraDetails)
from gateways.n_session import NSession
from gateways.the_gateway import TheGateway
from gateways.utils import GatewayUtils, Utils


class Colony:
    # pylint: disable=too-many-public-methods
    # pylint: disable=too-many-lines
    def __init__(self, account: str, **kwargs):
        """
        :param account: account name
        :param Keyword Arguments:
            * access_token: either access token, or long token
            * email: account email
            * password: account password
            * host: address of Colony host machine, including port. Example: 'http://192.168.30.22:5050'
            * space: current space to be used
            * api_version: api version to be used
        """
        # raising an error when not provided valid credentials
        if (not kwargs.get("access_token")) and not (kwargs.get("email") and kwargs.get("password")):
            credentials = {
                "access_token": kwargs.get("access_token"),
                "email": kwargs.get("email"),
                "password": kwargs.get("password")
            }
            raise UnauthorizedException(credentials)

        gateway = TheGateway(provider=kwargs.get("provider"), host=kwargs.get("host"))
        self.session = NSession()
        self.api_address = GatewayUtils.get_cs18_api_address(
            provider=kwargs.get("provider"), host=kwargs.get("host")
        ).lower()

        self.refresh_token = \
            kwargs.get("access_token") if kwargs.get("access_token") else \
                gateway.account_login(
                    account=account,
                    email=kwargs.get("email"),
                    password=kwargs.get("password")
                ).refresh_token

        # use either provided access_token, or login with provided credentials
        self.access_token = \
            kwargs.get("access_token") if kwargs.get("access_token") else \
                gateway.account_login(
                    account=account,
                    email=kwargs.get("email"),
                    password=kwargs.get("password")
                ).access_token

        self.session.add_header(
            "Authorization",
            "Bearer {}".format(self.access_token)
        )
        self.space = kwargs.get("space", "Trial")
        self._api_version = kwargs.get("api_version")
        self._api_root = ApiRoot(
            api_address=self.api_address, space=self.space, version=self._api_version
        )

    def change_space(self, space: str):
        self.space = space
        self._api_root = ApiRoot(
            api_address=self.api_address, space=self.space, version=self._api_version
        )
        return self

    def get_spaces(self) -> List[GetSpacesResponse]:
        method_url = self._api_root.spaces.spaces()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        spaces_json = json.loads(resp.text)
        return [Converters.create_spaces_response(space) for space in spaces_json]

    def get_account_status(self) -> AccountStatusResponse:
        method_url = self._api_root.settings.status()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return Converters.create_account_status_response(json.loads(resp.text))

    def get_about(self):
        method_url = self._api_root.about.api_about()
        resp = self.session.get(url=method_url)
        json_response = json.loads(resp.text)
        return json_response

    def remove_compute_service(self, cloud_account: str, compute_service: str):
        method_url = self._api_root.settings.compute_services()
        body = {"compute_service_name": f'{cloud_account}/{compute_service}'}
        res = self.session.delete(url=method_url, json=body)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_cloud_accounts_under_account(self) -> List[CloudAccountResponse]:
        method_url = self._api_root.settings.cloud_accounts()
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        json_response = json.loads(resp.text)
        cloud_accounts = [
            Converters.create_cloud_account_response(x) for x in json_response
        ]
        return cloud_accounts

    def register_azure_account(self, with_tags: bool = False, **kwargs):
        """
        :param with_tags: bool
        :param kwargs:  name: str
                        subscription_id: str,
                        tenant_id: str,
                        application_id: str,
                        application_secret: str,
                        management_resource_group: str,
                        external_key: [str]
        """
        params = {"add_tags": with_tags}
        method_url = self._api_root.settings.azure_cloud_accounts()
        request = {
            "name": kwargs.get("name"),
            "subscription_id": kwargs.get("subscription_id"),
            "tenant_id": kwargs.get("tenant_id"),
            "application_id": kwargs.get("application_id"),
            "application_secret": kwargs.get("application_secret"),
            "management_resource_group": kwargs.get("management_resource_group"),
        }
        if kwargs.get("external_key"):
            request["id"] = kwargs.get("external_key")

        resp = self.session.post(url=method_url, json=request, params=params)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def register_aws_account(
            self,
            arn_role: str,
            external_id: str,
            name: str = None,
            external_key: str = None,
    ):
        """ Register AWS
        :param arn_role: IAM role ARN required for assume role to the customer cloud account
        :param external_id: External ID required for assume role to the customer cloud account
        :param name: Cloud account given name. If not passed its automatically generated by some pattern.
        :param external_key: Predefined external key that was already generated for another cloud account.
        May be used only for CI, tests and developers environments. Should not be used in production
        """
        method_url = self._api_root.settings.aws_cloud_accounts()
        request = {"arn_role": arn_role, "external_id": external_id}

        if arn_role is not None:
            print("arn_role " + arn_role)

        if name:
            request["name"] = name
        if external_key:
            request["id"] = external_key

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def register_gcp_stub_account(self, name: str):
        method_url = self._api_root.settings.gcp_stub_cloud_accounts()
        request = {
            "name": name
        }
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    # todo old function -> tomer delete this in phase 2
    def add_azure_aks(self,
                      name: str,
                      cloud_account_name: str,
                      kube_config: str,
                      ingress_controller_type: str = None,
                      ingress_class: str = None,
                      configure_dns: bool = False,
                      generate_certificate: bool = False):
        """ Register AKS"""
        method_url = self._api_root.settings.add_compute_service_aks()
        request = {
            "kube_config": kube_config,
            "name": name,
            "cloud_account_name": cloud_account_name,
            "ingress_controller_type": ingress_controller_type,
            "ingress_class": ingress_class,
            "configure_dns": configure_dns,
            "generate_certificate": generate_certificate
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    # todo old function -> tomer delete this in phase 2
    def add_aws_k8s_unmanaged(self,
                              name: str,
                              cloud_account_name: str,
                              kube_config: str,
                              ingress_controller_type: str = None,
                              ingress_class: str = None,
                              configure_dns: bool = False,
                              generate_certificate: bool = False):
        method_url = self._api_root.settings.add_compute_service_aws_k8s_unmanaged()
        request = {
            "kube_config": kube_config,
            "name": name,
            "cloud_account_name": cloud_account_name,
            "ingress_controller_type": ingress_controller_type,
            "ingress_class": ingress_class,
            "configure_dns": configure_dns,
            "generate_certificate": generate_certificate
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    # new api
    def add_k8s_compute_service(self,
                                compute_service_name: str,
                                cloud_account_name: str,
                                agent_namespace: str,
                                ingress_controller_type: str = None,
                                ingress_class: str = None,
                                configure_dns: bool = False,
                                generate_certificate: bool = False):
        method_url = self._api_root.settings.add_k8s_compute_service_under_cloud_account()
        request = {
            "compute_service_name": compute_service_name,
            "cloud_account_name": cloud_account_name,
            "agent_namespace": agent_namespace,
            "configure_dns": configure_dns,
            "generate_certificate": generate_certificate
        }
        if ingress_controller_type:
            request["ingress_controller_type"] = ingress_controller_type
        if ingress_class:
            request["ingress_class"] = ingress_class

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_k8s_agent_deployment_yaml(self, cloud_account_name: str, compute_service_name: str,
                                      sandbox_namespaces: List[str]):
        method_url = self._api_root.settings.k8s_agent_deployment_yaml(cloud_account_name=cloud_account_name,
                                                                       compute_service_name=compute_service_name,
                                                                       sandbox_namespaces=','.join(sandbox_namespaces))
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp.text

    def get_k8s_cluster_agents(self, cluster_id: str):
        method_url = self._api_root.agent_connector.get_k8s_cluster_agents(cluster_id=cluster_id)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp.text

    def rename_cloud_account(self, old_name: str, new_name: str):
        method_url = self._api_root.settings.rename_cloud_accounts(old_name, new_name)

        resp = self.session.put(url=method_url)
        GatewayUtils.handle_response(resp, return_codes=[200])

    def get_account_template_url(self, external_id: str):
        """ Register AWS
        :param external_id: External ID required for assume role to the customer cloud account"""

        method_url = "{}?external_id={}".format(
            self._api_root.settings.aws_template(), external_id
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        template_json = json.loads(resp.text)
        return template_json

    def unregister_cloud_account(self, cloud_account_name: str):
        method_url = self._api_root.settings.cloud_accounts()
        body = {"cloud_account_name": cloud_account_name}
        resp = self.session.delete(url=method_url, json=body)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_aws_s3_artifacts_repo(self, account_name: str, bucket_name: str):
        method_url = self._api_root.artifacts_repository.aws_s3_storage(
            space_name=self.space
        )
        request = {"account_name": account_name, "bucket_name": bucket_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_azure_artifact_repository(self, account_name: str, storage_name: str):
        method_url = self._api_root.artifacts_repository.azure_storage(
            space_name=self.space
        )
        request = {"account_name": account_name, "storage_name": storage_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_artifact_repository(self):
        method_url = self._api_root.artifacts_repository.artifact_repos(
            space_name=self.space
        )
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return Converters.create_artifact_repo_response(json.loads(res.text))

    def get_artifact_repo_storage(self, cloud_account: str):
        method_url = self._api_root.artifacts_repository.artifact_repos_by_cloud_account_name(
            cloud_account_name=cloud_account, space_name=self.space
        )
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return json.loads(res.text)

    def remove_artifact_repository(self):
        method_url = self._api_root.artifacts_repository.artifact_repos(
            space_name=self.space
        )
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def create_long_token(self) -> TokenResponse:
        method_url = self._api_root.token.long_token()
        resp = self.session.post(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        token_json = json.loads(resp.text)
        return Converters.create_token_response(token_json)

    def revoke_long_token(self, long_token: str):
        method_url = self._api_root.token.revoke_long_token()
        self.session.add_header("Authorization", "Bearer {}".format(long_token))
        resp = self.session.post(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_blueprints(
            self, exclude_samples=False, deployment_type: str = None
    ) -> List[BlueprintResponse]:
        params = {}
        if deployment_type:
            params["type"] = deployment_type
        method_url = self._api_root.blueprint.blueprints_by_space_name(self.space)
        resp = self.session.get(url=method_url, params=params)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprints_json = json.loads(resp.text)
        all_blueprints = [
            Converters.create_blueprint_response(blueprint)
            for blueprint in blueprints_json
        ]
        if exclude_samples:
            all_blueprints = [
                blueprint for blueprint in all_blueprints if not blueprint.is_sample
            ]
        return all_blueprints

    def list_space_users(self) -> List[UserPermittedToSpaceResponse]:
        method_url = self._api_root.spaces.space_users(self.space)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        users_json = json.loads(resp.text)
        return [
            Converters.create_user_permitted_to_space_response(user)
            for user in users_json
        ]

    def publish_blueprint_in_catalog(self, blueprint_name: str):
        method_url = self._api_root.catalog.catalog(space_name=self.space)
        request = {"blueprint_name": blueprint_name}
        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_blueprint_files(
            self, blueprint_name: str, branch: str
    ) -> List[BlueprintFileResponse]:
        method_url = self._api_root.blueprint.blueprint_files(
            space_name=self.space, blueprint_name=blueprint_name, branch=branch
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        files = json.loads(resp.text)
        return [Converters.create_blueprint_file_response(file) for file in files]

    def list_blueprints_in_catalog(self) -> List[CatalogForGetAllResponse]:
        method_url = self._api_root.catalog.catalog(space_name=self.space)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprints_json = json.loads(resp.text)
        return [
            Converters.create_catalog_blueprint_response(blueprint)
            for blueprint in blueprints_json
        ]

    def unpublish_blueprint_in_catalog(self, blueprint_name: str):
        method_url = self._api_root.catalog.blueprint_in_catalog(
            space_name=self.space, blueprint_name=blueprint_name
        )
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_blueprint_details_in_catalog(
            self, blueprint_name: str
    ) -> CatalogForGetResponse:
        method_url = self._api_root.catalog.blueprint_in_catalog(
            space_name=self.space, blueprint_name=blueprint_name
        )
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        blueprint = json.loads(resp.text)
        return Converters.create_catalog_blueprint_full_response(blueprint)

    def list_sandboxes(self, retry: bool = True, **kwargs) -> List[SandboxResponseLean]:
        """
       :param retry: bool
       :param kwargs:  filter: str(automation/all/my),
                       sandbox_name: str,
                       count: int
       """

        method_url = self._api_root.sandbox.sandboxes()
        resp = self.session.get(url=method_url, params=kwargs)
        if resp.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.list_sandboxes(False, **kwargs)
            raise TimeoutError()
        GatewayUtils.handle_response(resp, return_codes=[200])
        result = json.loads(resp.text)
        return [
            Converters.create_sandbox_response_list_item(sandbox) for sandbox in result
        ]

    def list_productions(self, retry: bool = True) -> List[ProductionResponseLean]:
        url = self._api_root.production.productions()
        response = self.session.get(url=url)
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.list_productions(False)
            raise TimeoutError()
        GatewayUtils.handle_response(response, return_codes=[200])
        json_result = json.loads(response.text)
        return [
            Converters.create_production_response_list_item(item)
            for item in json_result
        ]

    def start_production(self, blueprint_name: str, artifacts: dict = None, inputs=None, compute_availability=None):
        url = self._api_root.production.productions()
        response = self.session.post(
            url=url,
            json={
                "blueprint_name": blueprint_name,
                "artifacts": artifacts if artifacts else {},
                "inputs": inputs if inputs else {},
                "compute_availability": compute_availability,
            },
        )
        GatewayUtils.handle_response(response=response, return_codes=[202])
        json_result = json.loads(response.text)
        return CreateProductionResponse(production_id=json_result["id"])

    def start_production_green(self,
                               production_id: str,
                               artifacts: dict = None,
                               inputs=None,
                               compute_availability: str = None):
        url = self._api_root.production.green_by_id(production_id=production_id)
        response = self.session.post(url=url, json={
            "artifacts": artifacts if artifacts else {},
            "inputs": inputs if inputs else {},
            "compute_availability": compute_availability
        })
        GatewayUtils.handle_response(response=response, return_codes=[202])
        json_result = json.loads(response.text)
        return CreateProductionResponse(production_id=json_result["id"])

    def update_production_debugging_service(self, production_id: str,
                                            environment: ProductionEnvironment,
                                            value: DebuggingServiceValue):
        if environment == ProductionEnvironment.BLUE:
            url = f'{self._api_root.production.blue_debugging_service(production_id=production_id)}'
        elif environment == ProductionEnvironment.GREEN:
            url = f'{self._api_root.production.green_debugging_service(production_id=production_id)}'
        else:
            raise ValueError(f'Unknown environment "{environment}"')

        response = self.session.put(url=f'{url}?value={value.value}')
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def expose_green(self, production_id: str, exposure_value: int):
        url = self._api_root.production.expose_green(production_id=production_id, exposure_value=exposure_value)
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def promote_green(self, production_id: str):
        url = self._api_root.production.promote_green(production_id=production_id)
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def start_sandbox(
            self, sandbox_name: str, blueprint_name: str, **kwargs
    ) -> CreateSandboxResponse:
        """
        :param sandbox_name: str
        :param blueprint_name: str
        :param kwargs:  artifacts: dict,
                        inputs: dict,
                        scheduled_end_time: datetime,
                        automation: bool,
                        duration: timedelta,
                        compute_availability: str,
                        lazy_load_artifacts: bool,
                        lazy_load_artifacts_timeout: int,
                        visibility: str
        """
        method_url = self._api_root.sandbox.sandboxes()
        request = {
            "blueprint_name": blueprint_name,
            "sandbox_name": sandbox_name,
            "artifacts": kwargs.get("artifacts", {}),
            "inputs": kwargs.get("inputs", {}),
            "automation": kwargs.get("automation", False),
            "compute_availability": kwargs.get("compute_availability"),
            "lazy_load_artifacts": kwargs.get("lazy_load_artifacts", False),
            "lazy_load_artifacts_timeout": kwargs.get("lazy_load_artifacts_timeout", 0),
            "visibility": kwargs.get("visibility")
        }

        if kwargs.get("scheduled_end_time"):
            request["scheduled_end_time"] = datetime.isoformat(kwargs.get("scheduled_end_time"))

        if kwargs.get("duration"):
            request["duration"] = duration_isoformat(kwargs.get("duration"))

        if "branch" in kwargs or "commit" in kwargs:
            request["source"] = {
                "branch": kwargs.get("branch"),
                "commit": kwargs.get("commit"),
            }

        resp = self.session.post(
            url=method_url,
            json=request,
        )
        GatewayUtils.handle_response(response=resp, return_codes=[202])
        result = json.loads(resp.text)
        return CreateSandboxResponse(sandbox_id=result["id"])

    def end_sandbox(
            self,
            sandbox_id: str,
            retry: bool = True
    ):
        method_url = self._api_root.sandbox.sandbox_by_id(sandbox_id=sandbox_id)
        resp = self.session.delete(url=method_url)

        if resp.status_code in [408, 429, 504]:
            if retry:
                time.sleep(30)  # need to make sure that the next call will not use the cached timedout error
                self.end_sandbox(sandbox_id, False)
                return

        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def end_production(self, production_id: str):
        method_url = self._api_root.production.production_by_id(
            production_id=production_id
        )
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def end_production_green(self, production_id: str):
        method_url = self._api_root.production.green_by_id(production_id=production_id)
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[202])

    def get_sandbox_details(
            self, sandbox_id: str, retry: bool = True
    ) -> Optional[SandboxResponse]:
        method_url = self._api_root.sandbox.sandbox_by_id(sandbox_id=sandbox_id)
        resp = self.session.get(url=method_url)
        if resp.status_code == 404:
            return None
        if resp.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_sandbox_details(sandbox_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=resp, return_codes=[200])
        result = json.loads(resp.text)
        sandbox_response = Converters.create_sandbox_response(result)
        return sandbox_response

    def get_production_details(
            self, production_id: str, retry: bool = True) \
            -> Optional[ProductionBlueResponse]:
        url = self._api_root.production.production_by_id(production_id)
        response = self.session.get(url=url)
        if response.status_code == 404:
            return None
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_production_details(production_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        production_response = Converters.create_production_blue_response(json_result)
        return production_response

    def get_production_green_details(
            self,
            production_id: str,
            retry: bool = True
    ) -> Optional[ProductionGreenResponse]:
        url = self._api_root.production.green_by_id(production_id)
        response = self.session.get(url=url)
        if response.status_code == 404:
            return None
        if response.status_code in [408, 429]:
            if retry:
                time.sleep(30)
                return self.get_production_green_details(production_id, False)
            raise TimeoutError()

        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        production_response = Converters.create_production_green_response(json_result)
        return production_response

    def add_account_blueprint_repository_using_token(self, request: AddBlueprintUsingTokenRepositoryRequest):
        method_url = self._api_root.settings.add_repository_by_token()
        resp = self.session.post(url=method_url, json=request.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_account_blueprint_github_repository(self, request: AddBlueprintGithubRepositoryRequest):
        method_url = self._api_root.settings.add_github_repository()
        resp = self.session.post(url=method_url, json=request.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_account_blueprint_bitbucket_repository(self, request: AddBlueprintBitbucketRepositoryRequest):
        method_url = self._api_root.settings.add_bitbucket_repository()
        resp = self.session.post(url=method_url, json=request.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def delete_account_blueprint_repository(self, name: str):
        method_url = self._api_root.settings.repository(name)
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_account_blueprint_repositories(self) -> List[AccountRepositoryResponse]:
        url = self._api_root.settings.repositories()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

        repositories_json = json.loads(response.text)
        return [Converters.create_account_repository_response(repository) for repository in repositories_json]

    def get_account_blueprint_repository(self, repository_name: str) -> AccountRepositoryResponse:
        method_url = self._api_root.settings.repository(repository_name)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

        repository_json = json.loads(response.text)
        return Converters.create_account_repository_response(repository_json)

    def get_account_blueprint_repository_details(self, repository_name: str) -> AccountRepositoryDetailsResponse:
        method_url = self._api_root.settings.repository_details(repository_name)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

        repository_details_json = json.loads(response.text)
        return Converters.create_account_repository_details_response(repository_details_json)

    def add_repository_to_space(
            self,
            repository_details: BlueprintRepositoryDetails,
            deployment_type: str = None,
            wait_for_cache_refresh: bool = False,
    ):
        print(
            f'repository_url: {repository_details.repository_url} access_token: {repository_details.access_token}')
        method_url = self._api_root.spaces.space_repositories(space_name=self.space)
        request = {
            "repository_url": repository_details.repository_url,
            "access_token": repository_details.access_token,
            "repository_type": repository_details.repository_type,
            "branch": repository_details.branch,
            "type": deployment_type
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        if wait_for_cache_refresh:
            Utils.wait_for(
                func=lambda: (not self.list_blueprints(exclude_samples=True) is False),
                interval_sec=5,
                max_retries=24,
                subject="blueprints ready",
            )

    def add_bitbucket_repository_to_space_via_token(
            self,
            bitbucket_blueprint_repository_details: BitbucketBlueprintRepositoryDetails,
            deployment_type: str = None,
            wait_for_cache_refresh: bool = False,
    ):

        method_url = self._api_root.spaces.space_repositories(space_name=self.space, repository_type="bitbucket")
        request = {
            "repository_url": bitbucket_blueprint_repository_details.blueprint_repository_details.repository_url,
            "code": bitbucket_blueprint_repository_details.auth_code,
            "repository_type": bitbucket_blueprint_repository_details.blueprint_repository_details.repository_type,
            "branch": bitbucket_blueprint_repository_details.blueprint_repository_details.branch,
            "type": deployment_type,
            "redirection_url": bitbucket_blueprint_repository_details.redirect_url
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        if wait_for_cache_refresh:
            Utils.wait_for(
                func=lambda: (not self.list_blueprints(exclude_samples=True) is False),
                interval_sec=5,
                max_retries=24,
                subject="blueprints ready",
            )

    def remove_repository_from_space(self, wait_for_cache_refresh=False):
        method_url = self._api_root.spaces.space_repositories(space_name=self.space)
        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        if wait_for_cache_refresh:
            Utils.wait_for(
                func=lambda: (not self.list_blueprints(exclude_samples=True) is True),
                interval_sec=5,
                max_retries=24,
                subject="blueprints removed",
            )

    def get_tfstate_file(self, sandbox_id: str, service_name: str) -> bytes:
        method_url = self._api_root.sandbox.sandbox_tfstate(sandbox_id=sandbox_id, service_name=service_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp.content

    def validate_blueprint(self, blueprint: str, deployment_type: str = 'sandbox', branch: str = None,
                           commit: str = None) -> BlueprintValidationResponse:
        method_url = self._api_root.blueprint_validation.blueprint_validation_by_space_name(space_name=self.space)
        request = {"blueprint_name": blueprint,
                   "type": deployment_type}
        if branch or commit:
            request['source'] = {"branch": branch,
                                 "commit": commit}
        response = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        return Converters.create_blueprint_validation_response(json_result)

    def invite_user(self, invites: CreateInvitationsRequest):
        method_url = self._api_root.account.invitations()
        resp = self.session.post(url=method_url, json=invites.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_invites(self, space_name: str = None) -> List[UserInvitationResponse]:
        method_url = self._api_root.account.account_invitations(space=space_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        result = json.loads(resp.text)
        return [Converters.create_user_invitation_response(x) for x in result]

    def get_invited_user_details_by_secret(self, secret: str):
        method_url = self._api_root.account.account_signup_by_secret(secret=secret)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        result = json.loads(resp.text)
        return Converters.create_invited_user_details_response(result)

    def remove_invites(self, emails: List[str]):
        method_url = self._api_root.account.account_invitations()
        req = {"emails": emails}
        resp = self.session.delete(url=method_url, json=req)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_repositories(
            self, repository_type: str = None
    ) -> List[RepositoryResponse]:
        params = {}
        if repository_type:
            params["type"] = repository_type
        url = self._api_root.spaces.space_repositories(space_name=self.space)
        response = self.session.get(url=url, params=params)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        return [Converters.create_repository_response(x) for x in json_result]

    def update_scheduled_end_time(self, sandbox_id: str, scheduled_end_time: datetime):
        url = "{base}?value={end_time}".format(
            base=self._api_root.sandbox.sandbox_scheduled_end_time(
                sandbox_id=sandbox_id
            ),
            end_time=quote(scheduled_end_time.isoformat()),
        )
        response = self.session.put(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def turn_on_debugging_service(self, sandbox_id: str):
        self.update_debugging_service(sandbox_id=sandbox_id, value=DebuggingServiceValue.ON)

    def turn_off_debugging_service(self, sandbox_id: str):
        self.update_debugging_service(sandbox_id=sandbox_id, value=DebuggingServiceValue.OFF)

    def update_debugging_service(self, sandbox_id: str, value: DebuggingServiceValue):
        url = self._api_root.sandbox.sandbox_debugging_service(sandbox_id=sandbox_id)
        response = self.session.put(url=f'{url}?value={value.value}')
        GatewayUtils.handle_response(response=response, return_codes=[200, 202])

    def get_space_achievements(self) -> dict:
        url = self._api_root.achievements.achievements(space_name=self.space)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def create_space(self, space_name: str, cloud_accounts: List[str] = None):
        url = self._api_root.spaces.spaces()
        request = {"name": space_name, "cloud_accounts": cloud_accounts}

        response = self.session.post(url=url, json=request)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def delete_space(self, space_name: str):
        url = self._api_root.spaces.space_by_name(space_name=space_name)
        response = self.session.delete(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_space(self, name: str) -> GetSpaceResponse:
        url = self._api_root.spaces.space_by_name(space_name=name)
        res = self.session.get(url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return GetSpaceResponse(json.loads(res.text))

    def get_images(self, space_name: str, cloud_account_name: str, region_id: str):
        url = self._api_root.spaces.get_images(space_name, cloud_account_name, region_id)
        res = self.session.get(url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return [Converters.get_image_response(x) for x in json.loads(res.text)]

    def get_instance_types(self, space_name: str, cloud_account_name: str, region_id: str):
        url = self._api_root.spaces.get_instance_types(space_name, cloud_account_name, region_id)
        res = self.session.get(url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return [Converters.get_instance_type_response(x) for x in json.loads(res.text)]

    def update_space(self, space: str, update_space_request: UpdateSpaceRequest):
        url = self._api_root.spaces.space_by_name(space_name=space)
        response = self.session.put(url=url, json=update_space_request.__dict__)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_space_roles(self) -> List[RoleListItemResponse]:
        url = self._api_root.settings.space_roles()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        roles_json = json.loads(response.text)
        return [Converters.create_role_list_item_response(role) for role in roles_json]

    def get_account_roles(self) -> List[RoleListItemResponse]:
        url = self._api_root.settings.account_roles()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        roles_json = json.loads(response.text)
        return [Converters.create_role_list_item_response(role) for role in roles_json]

    def remove_cloud_account_from_space(self, cloud_account: str):
        method_url = self._api_root.spaces.space_cloud_account(
            self.space, cloud_account
        )
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_cloud_account_to_space(self, cloud_account: str,
                                   infra_settings: SpaceCloudAccountInfraSettingsRequest = None):
        method_url = self._api_root.spaces.space_cloud_accounts(self.space)
        data = {"name": cloud_account}
        if infra_settings:
            data['internet_facing'] = infra_settings.internet_facing
            data['existing_infra'] = infra_settings.existing_infra
            if infra_settings.existing_infra_sets:
                existing_infra_sets = []
                for infra_set in infra_settings.existing_infra_sets:
                    existing_infra_sets.append(
                        {
                            'region': infra_set.region,
                            'virtual_network': infra_set.virtual_network,
                            'subnets': {'gateway_subnet': infra_set.subnets.gateway_subnet,
                                        'management_subnet': infra_set.subnets.management_subnet,
                                        'application_subnets': infra_set.subnets.application_subnets
                                        } if infra_set.subnets else None
                        }
                    )
                data['existing_infra_sets'] = existing_infra_sets
        res = self.session.post(url=method_url, json=data)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_cloud_account_in_space(self, cloud_account: str, infra_settings: SpaceCloudAccountInfraSettingsRequest):
        method_url = self._api_root.spaces.space_cloud_account(space_name=self.space, cloud_account=cloud_account)
        data = {
            'internet_facing': infra_settings.internet_facing,
            'existing_infra': infra_settings.existing_infra
        }
        if infra_settings.existing_infra_sets:
            existing_infra_sets = []
            for infra_set in infra_settings.existing_infra_sets:
                existing_infra_sets.append(
                    {
                        'region': infra_set.region,
                        'virtual_network': infra_set.virtual_network,
                        'subnets': {'gateway_subnet': infra_set.subnets.gateway_subnet,
                                    'management_subnet': infra_set.subnets.management_subnet,
                                    'application_subnets': infra_set.subnets.application_subnets
                                    } if infra_set.subnets else None
                    }
                )
            data['existing_infra_sets'] = existing_infra_sets
        res = self.session.put(url=method_url, json=data)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_cloud_accounts_under_space(self) -> List[CloudAccountInSpaceResponse]:
        method_url = self._api_root.spaces.space_cloud_accounts(self.space)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return [
            Converters.create_cloud_account_in_space_response(x) for x in json.loads(res.text)
        ]

    def add_compute_service_to_space(self, cloud_account: str, request: AddK8SComputeServiceToSpaceRequest):
        method_url = self._api_root.spaces.space_cloud_account_compute_services(space_name=self.space,
                                                                                cloud_account=cloud_account)
        res = self.session.post(url=method_url, json=request.__dict__)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_compute_service_in_space(self, cloud_account: str, compute_service: str,
                                        request: UpdateK8SComputeServiceInSpaceRequest):
        method_url = self._api_root.spaces.space_cloud_account_compute_service(space_name=self.space,
                                                                               cloud_account=cloud_account,
                                                                               compute_service=compute_service)
        res = self.session.put(url=method_url, json=request.__dict__)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def remove_compute_service_from_space(self, cloud_account: str, compute_service: str):
        method_url = self._api_root.spaces.space_cloud_account_compute_service(space_name=self.space,
                                                                               cloud_account=cloud_account,
                                                                               compute_service=compute_service)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_account_users(self) -> List[UserForAllUsersResponse]:
        method_url = self._api_root.account.account_users()
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        users_json = json.loads(res.text)
        return [Converters.create_user_for_all_users_response(x) for x in users_json]

    def get_space_permissions(self) -> List[str]:
        method_url = self._api_root.spaces.user_permissions(self.space)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return json.loads(res.text)

    def delete_account(self, account_name: str):
        method_url = self._api_root.account.delete_account(account_name)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        return res.text

    def contact_me(self, first_name: str, last_name: str, email: str, phone: str):
        method_url = self._api_root.account.contact_me()
        request = {
            "first_name": first_name,
            "last_name": last_name,
            "email": email,
            "phone": phone
        }
        res = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_user_to_space(self, user_email: str, space_role: str):
        method_url = self._api_root.spaces.space_users(self.space)
        request = {"email": user_email, "space_role": space_role}
        res = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_user_space_role(self, user_email: str, space_role: str):
        method_url = self._api_root.spaces.user_space_role(self.space, user_email, space_role)
        res = self.session.put(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def remove_user_from_space(self, user_email: str):
        method_url = self._api_root.spaces.space_user(self.space, user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def update_user_account_role(self, user_email: str, account_role: str):
        method_url = self._api_root.account.account_user_account_role_value(user_email, account_role)
        res = self.session.put(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def remove_user_account_role(self, user_email: str):
        method_url = self._api_root.account.account_user_account_role(user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def get_access_url_by_link(self, access_link: AccessLink) -> str:
        method_url = f"{self.api_address}{access_link.link}"
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        res_json = json.loads(res.text)
        return res_json['url']

    def get_access_url_by_params(self, sandbox_id: str, instance_id: str, protocol: str) -> str:
        method_url = self._api_root.qualiy.connect(sandbox_id=sandbox_id, instance_id=instance_id, protocol=protocol)
        res = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        res_json = json.loads(res.text)
        return res_json['url']

    def delete_user(self, user_email: str):
        method_url = self._api_root.account.account_user(user_email)
        res = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])

    def add_artifactory_repo(self, server_url: str, artifactory_username: str, api_key: str):
        method_url = self._api_root.artifacts_repository.artifactory(space_name=self.space)
        request = {
            "server_url": server_url,
            "artifactory_username": artifactory_username,
            "api_key": api_key
        }
        response = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_cost_breakdown(self, frm: datetime, until: datetime, criteria: str = 'blueprint') -> dict:
        url = self._api_root.cost.breakdown(space_name=self.space, frm=frm, until=until, criteria=criteria)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def get_cost_usage(self, date: datetime, timezone: str = '') -> dict:
        url = self._api_root.cost.usage(space_name=self.space, date=date, timezone=timezone)
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response, [200])
        return json.loads(response.text)

    def post_request(self, url: str, request_json: dict, return_codes: List = None):
        return_codes = return_codes or [200]
        response = self.session.post(url=url, json=request_json)
        GatewayUtils.handle_response(response=response, return_codes=return_codes)

    def put_request(self, url: str, request_json: dict, return_codes: List = None):
        return_codes = return_codes or [200]
        response = self.session.put(url=url, json=request_json)
        GatewayUtils.handle_response(response=response, return_codes=return_codes)

    def set_account_extra_details(self, account_name: str, sso_enabled: bool, anonymous_users: bool,
                                  sso_update_enabled: bool):
        set_account_extra_details_url = self._api_root.account.set_account_extra_details(account_name)
        self.put_request(set_account_extra_details_url, {"sso_enabled": sso_enabled, "anonymous_users": anonymous_users,
                                                         "sso_update_enabled": sso_update_enabled}
                         , return_codes=[200])

    def get_account_extra_details(self, account_name: str) -> AccountExtraDetails:
        set_account_extra_details_url = self._api_root.account.set_account_extra_details(account_name)
        response = self.session.get(url=set_account_extra_details_url)
        GatewayUtils.handle_response(response, [200])
        return Converters.create_account_extra_details(json=json.loads(response.text))

    def get_space_cloud_account_regions(self, space_name: str, cloud_account_name: str) -> List[RegionResponse]:
        method_url = self._api_root.spaces.space_cloud_account_regions(space_name=space_name,
                                                                       cloud_account_name=cloud_account_name)
        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        regions_json = json.loads(resp.text)
        return [Converters.create_space_cloud_account_regions_response(region) for region in regions_json]

    def verify_cloud_account(self, cloud_account_name: str) -> VerifyCloudProviderResponse:
        method_url = self._api_root.settings.verify_cloud_account(cloud_account_name)
        res = self.session.put(url=method_url)
        GatewayUtils.handle_response(response=res, return_codes=[200])
        verify_json = json.loads(res.text)
        return Converters.create_verify_response(verify_json)

    def get_space_sandbox_repository(self, space_name: str):
        method_url = self._api_root.spaces.space_sandbox_repository(space_name)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 204])
        if not response.text:
            return None
        response_json = json.loads(response.text)
        return Converters.create_space_repository_response(response_json)

    def get_space_production_repository(self, space_name: str):
        method_url = self._api_root.spaces.space_production_repository(space_name)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 204])
        if not response.text:
            return None
        response_json = json.loads(response.text)
        return Converters.create_space_repository_response(response_json)

    def get_debugging_pem_file_content(self, space_name: str, sandbox_id: str, instance_id: str) -> str:
        method_url = self._api_root.debugging.get_aws_pem_file(space_name, sandbox_id, instance_id)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        return response.text

    def get_direct_aws_ssh_connect(self, space_name: str, sandbox_id: str, instance_id: str) -> DirectAwsSshResponse:
        method_url = self._api_root.debugging.get_direct_aws_ssh_connect(space_name, sandbox_id, instance_id)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        response_json = json.loads(response.text)
        return Converters.create_direct_aws_ssh_response(response_json)

    def get_direct_azure_ssh_connect(self, space_name: str, sandbox_id: str, instance_id: str) \
            -> DirectAzureSshResponse:
        method_url = self._api_root.debugging.get_direct_azure_ssh_connect(space_name, sandbox_id, instance_id)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        response_json = json.loads(response.text)
        return Converters.create_direct_azure_ssh_response(response_json)

    def get_debugging_rdp_file_content(self, space_name: str, sandbox_id: str, instance_id: str) -> str:
        method_url = self._api_root.debugging.get_direct_rpd_file(space_name, sandbox_id, instance_id)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        return response.text

    def get_direct_rdp_connect(self, space_name: str, sandbox_id: str, instance_id: str) -> DirectRdpResponse:
        method_url = self._api_root.debugging.get_direct_rpd_connect(space_name, sandbox_id, instance_id)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        response_json = json.loads(response.text)
        return Converters.create_direct_rdp_response(response_json)

    def add_parameter_store(self, parameter: ParameterStoreItem):
        method_url = self._api_root.parameterStore.parameters()
        response = self.session.post(url=method_url, json=Utils.get_as_json(parameter))
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_all_parameters(self):
        method_url = self._api_root.parameterStore.parameters()
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 204])
        response_json = json.loads(response.text)
        return Converters.create_get_all_parameters_response(response_json)

    def get_parameter(self, parameter_name: str):
        method_url = self._api_root.parameterStore.parameter(parameter_name=parameter_name)
        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200, 204])
        response_json = json.loads(response.text)
        return Converters.create_get_parameter_response(response_json)

    def delete_parameter(self, parameter_name: str):
        method_url = self._api_root.parameterStore.parameter(parameter_name=parameter_name)
        response = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def set_aws_cloud_provider_settings(self, model: AwsCloudProviderSettingsModel):
        method_url = self._api_root.settings.aws_cloud_providers()

        response = self.session.put(url=method_url, json=model.__dict__)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_aws_cloud_provider_settings(self) -> AwsCloudProviderSettingsModel:
        method_url = self._api_root.settings.aws_cloud_providers()

        response = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        response_json = json.loads(response.text)
        return Converters.aws_cloud_provider_settings_model(data=response_json)

    def add_repository_provider_to_account(
            self,
            provider_name: str,
            repository=None,
            repo_relationship=None
    ):
        method_url = self._api_root.settings.repository_providers()
        request = {
            "name": provider_name,
            "type": "github",
        }

        if repository:
            repos = [{
                "name": repository['repository_name'],
                "allow_sharing": True,
                "open_access": True,
                "repository_url": repository['repository_details'].repository_url,
                "connection_details": {
                    "token": repository['repository_details'].access_token
                },
                "repository_relationships": None
            }]

            request["repos"]: List[dict] = repos

            if repo_relationship:
                repository_relationships = [{
                    "deployment_type": repo_relationship['deployment_type'],
                    "account_repository_name": repo_relationship['repository_name'],
                    "branch": repo_relationship['branch'],
                    "space_name": repo_relationship['space_name'],
                }]

                request["repos"][0]["repository_relationships"]: List[dict] = repository_relationships

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        return resp.text.replace('"', '')

    def get_repository_provider(
            self,
            provider_name: str
    ) -> RepositoryProviderResponse:
        method_url = self._api_root.settings.repository_provider(provider_name)

        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        json_result = json.loads(resp.text)
        return Converters.create_repository_provider_response(json_result)

    def get_repository_providers(
            self,
    ) -> List[RepositoryProviderResponse]:
        method_url = self._api_root.settings.repository_providers()

        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        json_result = json.loads(resp.text)
        return [Converters.create_repository_provider_response(provider) for provider in json_result]

    def delete_repository_provider(
            self,
            provider_name: str
    ):
        method_url = self._api_root.settings.repository_provider(provider_name)

        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def add_repository_to_account(
            self,
            repository_details: BlueprintRepositoryDetails,
            repository_provider_id: str,
            repository_name: str,
            repo_relationship: dict = None

    ):
        method_url = self._api_root.settings.repositories()

        request = {
            "provider_id": repository_provider_id,
            "name": repository_name,
            "allow_sharing": True,
            "open_access": True,
            "repository_url": repository_details.repository_url,
            "connection_details": {
                "token": repository_details.access_token
            },
        }

        if repo_relationship:
            repository_relationships = [{
                "deployment_type": repo_relationship['deployment_type'],
                "account_repository_name": repo_relationship['repository_name'],
                "branch": repo_relationship['branch'],
                "space_name": repo_relationship['space_name'],
            }]

            request["repository_relationships"]: List[dict] = repository_relationships

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def list_account_repositories(self) -> List[AccountRepositoryResponse]:
        url = self._api_root.settings.repositories()
        response = self.session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        json_result = json.loads(response.text)
        return [Converters.create_account_repository_response(x) for x in json_result]

    def delete_account_repository(self, name: str):
        url = self._api_root.settings.repository(name)
        response = self.session.delete(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def add_repository_relationship_to_space(
            self,
            deployment_type: str,
            account_repository_name: str,
            space: str):

        self.__add_repository_relationship(deployment_type, account_repository_name, space, None)

    def add_repository_relationship(
            self,
            deployment_type: str,
            account_repository_name: str,
            branch: str = None):

        self.__add_repository_relationship(deployment_type, account_repository_name, self.space, branch)

    def __add_repository_relationship(
            self,
            deployment_type: str,
            account_repository_name: str,
            space: str,
            branch: str):

        method_url = self._api_root.spaces.add_repository_relationship(space)

        request = {
            "deployment_type": deployment_type,
            "account_repository_name": account_repository_name,
            "branch": branch
        }

        resp = self.session.post(url=method_url, json=request)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def get_repository_relationship(
            self,
            deployment_type: str
    ) -> RepositoryRelationshipResponse:
        method_url = self._api_root.spaces.repository_relationship(self.space, deployment_type)

        resp = self.session.get(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])
        json_result = json.loads(resp.text)
        return Converters.create_repository_relationship_response(json_result)

    def delete_repository_relationship(
            self,
            deployment_type: str,
    ):
        self.delete_repository_relationship_from_space(deployment_type, self.space)

    def delete_repository_relationship_from_space(
            self,
            deployment_type: str,
            space: str
    ):
        method_url = self._api_root.spaces.repository_relationship(space, deployment_type)

        resp = self.session.delete(url=method_url)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

    def submit_new_user_survey(self, survey: NewUserSurveyRequest) -> Response:
        method_url = self._api_root.account.submit_new_user_survey()
        resp = self.session.post(url=method_url, json=survey.__dict__)
        GatewayUtils.handle_response(response=resp, return_codes=[200])

        return resp


class SandboxesHandler:
    def __init__(self):
        self.sandboxes_set = set()

    def add_sandbox(self, sandbox_id):
        self.sandboxes_set.add(sandbox_id)
        print("Sandbox {id} added to cleaner".format(id=sandbox_id))

    def remove_sandbox(self, sandbox_id):
        self.sandboxes_set.remove(sandbox_id)
        print("Sandbox {id} removed from cleaner".format(id=sandbox_id))


class SandboxCleaner:
    def __init__(self, client: Colony, end_on_exception=True):
        self.end_on_exception = end_on_exception
        self.client = client
        self.sandboxes = SandboxesHandler()
        self.space = self.client.space

    def __enter__(self):
        return self.sandboxes

    def __exit__(self, exc_type, exc_val: str, exc_tb):
        for sandbox in self.sandboxes.sandboxes_set:
            # if exc_type is MaxRetriesException:
            #     print(f"Sandbox {sandbox} failed on timeout")
            if exc_type is SandboxNotFound:
                print(f"Sandbox {sandbox} failed on not found")
            elif exc_type is SandboxEndingFailed:
                print(f"Sandbox {sandbox} failed to end")
            else:
                if self.end_on_exception:
                    print("Sandbox {} is being removed by cleaner".format(sandbox))
                    self.client.end_sandbox(sandbox_id=sandbox)


class ProductionCleaner:
    def __init__(self, client: Colony):
        self.client = client
        self.ids = []
        self.space = self.client.space

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for production_id in self.ids:
            print("Production {} is being removed by cleaner".format(production_id))

            self.client.end_production(production_id=production_id)

    def add(self, production_id):
        self.ids.append(production_id)
        print("Production {id} added to cleaner".format(id=production_id))

    def remove(self, production_id):
        self.ids.remove(production_id)
        print("Production {id} removed from cleaner".format(id=production_id))



from gateways.apis.api_base_class import ApiBase


class ParameterStore(ApiBase):
    def parameters(self):
        return self.build_route("settings/parameters/")

    def parameter(self, parameter_name: str):
        return self.build_route(
            "settings/parameters/{parameter_name}".format(**locals()))



"""
Container for object classes used in cs18-api-client
"""
import abc

import dateutil.parser


class AccessLink:
    def __init__(self, protocol: str, link: str):
        self.link = link
        self.protocol = protocol


class Commit:
    def __init__(self, data: dict):
        self.sha = data["sha"]
        self.author = data["commit"]["author"]["name"]
        self.date = dateutil.parser.parse(data["commit"]["author"]["date"])
        self.comment = data["commit"]["message"]

    def __str__(self):
        return "{0}: [{1}] {2}".format(self.date, self.sha[:7], self.comment)


class ColonyAccount:
    def __init__(
            self, account: str, email: str, password: str, first_name: str, last_name: str
    ):
        self.account = account
        self.default_space = "Trial"
        self.sample_space = "Sample"
        self.email = email
        self.password = password
        self.first_name = first_name
        self.last_name = last_name


class BlueprintRepositoryDetails:
    def __init__(self, repository_url: str, access_token: str, repository_type: str, branch: str = None):
        self.repository_url = repository_url
        self.repository_type = repository_type
        self.access_token = access_token
        self.branch = branch


class BitbucketBlueprintRepositoryDetails:
    def __init__(self, auth_code: str, redirect_url: str, blueprint_repository_details: BlueprintRepositoryDetails):
        self.blueprint_repository_details = blueprint_repository_details
        self.auth_code = auth_code
        self.redirect_url = redirect_url


class AddAccountBlueprintRepositoryRequest(abc.ABC):
    def __init__(self, name: str, repository_url: str, allow_sharing: bool,
                 open_access: bool):
        self.name = name
        self.repository_url = repository_url
        self.open_access = open_access
        self.allow_sharing = allow_sharing


class AddBlueprintUsingTokenRepositoryRequest(AddAccountBlueprintRepositoryRequest):
    def __init__(self, name: str, repository_url: str, allow_sharing: bool,
                 open_access: bool, access_token: str, repository_type: str):
        super().__init__(name, repository_url, allow_sharing, open_access)
        self.repository_type = repository_type
        self.access_token = access_token


class AddBlueprintGithubRepositoryRequest(AddAccountBlueprintRepositoryRequest):
    def __init__(self, name: str, repository_url: str, allow_sharing: bool,
                 open_access: bool, code: str, state: str):
        super().__init__(name, repository_url, allow_sharing, open_access)
        self.code = code
        self.state = state


class AddBlueprintBitbucketRepositoryRequest(AddAccountBlueprintRepositoryRequest):
    def __init__(self, name: str, repository_url: str, allow_sharing: bool,
                 open_access: bool, code: str, redirection_url: str):
        super().__init__(name, repository_url, allow_sharing, open_access)
        self.code = code
        self.redirection_url = redirection_url



from datetime import datetime

from gateways.apis.api_base_class import ApiBase


class ApiCost(ApiBase):
    def breakdown(self, space_name: str, frm: datetime, until: datetime, criteria: str = 'blueprint'):
        return self.build_route(
            "spaces/{space_name}/cost/breakdown?criteria={criteria}&from={frm}&to={until}".format(**locals()))

    def usage(self, space_name: str, date: datetime, timezone: str = ''):
        return self.build_route("spaces/{space_name}/cost/usage?date={date}&timezone={timezone}".format(**locals()))



from gateways.apis.api_base_class import ApiBase


class ApiAchievements(ApiBase):
    def achievements(self, space_name: str):
        return self.build_route("spaces/{space_name}/achievements".format(**locals()))



from gateways.apis.api_base_class import ApiBase


class ApiBlueprint(ApiBase):
    def blueprints_by_space_name(self, space_name: str):
        return self.build_route("spaces/{space_name}/blueprints".format(**locals()))

    def blueprint_files(self, space_name: str, blueprint_name: str, branch: str):
        return self.build_route(
            "spaces/{space_name}/blueprints/{blueprint_name}/{branch}/files".format(
                **locals()
            )
        )



import json
from typing import Dict

from gateways.common.cs18_api_requests import UserSignupRequest, CreateAccountRequest
from gateways.common.cs18_api_responses import TokenResponse
from gateways.n_session import NSession
from gateways.utils import GatewayUtils
from gateways.common.cs18_api_converters import Converters


class TheGateway:
    """
    Colony gateway to preform actions that don't require previous authorization
    """

    def __init__(self, provider: str = None, host: str = None):
        self._session = NSession()
        self.api_address = GatewayUtils.get_cs18_api_address(
            provider=provider, host=host
        )

    def register_colony_account(self, req: CreateAccountRequest):
        response = self._session.post(
            url="{}/api/accounts/register".format(self.api_address),
            json={
                "account_name": req.account_name,
                "first_name": req.first_name,
                "last_name": req.last_name,
                "email": req.email,
                "password": req.password,
                "phone_number": req.phone_number,
            },
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def account_request_colony_account(self, req: CreateAccountRequest):
        response = self._session.post(
            url="{}/api/accounts/account_request".format(self.api_address),
            json={
                "account_name": req.account_name,
                "first_name": req.first_name,
                "last_name": req.last_name,
                "email": req.email,
                "password": req.password,
                "phone_number": req.phone_number,
                "utm_campaign": req.utm_campaign,
            },
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def account_finalize_colony_account(self, token):
        response = self._session.post(
            url="{}/api/accounts/account_finalize".format(self.api_address),
            json={
                "secret": token
            },
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def account_login(self, account: str, email: str, password: str) -> TokenResponse:
        response = self._session.post(
            url="{}/api/accounts/{}/login".format(self.api_address, account),
            json={"email": email, "password": password},
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])
        token_json = json.loads(response.text)
        return Converters.create_token_response(token_json)

    def signup_user(self, signup: UserSignupRequest):
        """
        Create a user signup action.
        Sends out an email to the invitee[s], that upon completion are added under the inviting Colony account.
        :param signup: UserSignupRequest
        :return:
        """
        response = self._session.post(
            url="{}/api/accounts/signup".format(self.api_address),
            json={
                "first_name": signup.first_name,
                "last_name": signup.last_name,
                "password": signup.password,
                "secret": signup.secret,
            },
        )

        GatewayUtils.handle_response(response=response, return_codes=[200])

    def token_refresh(self, access_token: str, refresh_token: str) -> TokenResponse:
        url = "{api_address}/api/token/refresh".format(
            api_address=self.api_address
        )
        data = {'refresh_token': refresh_token}
        response = self._session.post(
            url=url, json=data, headers=self._get_headers(access_token=access_token)
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])
        token_json = json.loads(response.text)
        return Converters.create_token_response(token_json)

    def token_revoke(self, access_token: str, refresh_token: str = None):
        url = "{api_address}/api/token/revoke".format(api_address=self.api_address)
        refresh_token_model = {"refresh_token": refresh_token}
        response = self._session.post(
            url=url, headers=self._get_headers(access_token=access_token),
            json=refresh_token_model
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_access_token_with_refresh_token(self, access_token: str, refresh_token: str):
        url = f"{self.api_address}/api/token/refresh".format(**locals())
        data = {'refresh_token': refresh_token}
        response = self._session.post(
            url=url, headers=self._get_headers(access_token=access_token), json=data
        )
        GatewayUtils.handle_response(response=response, return_codes=[200])

    def get_swagger_doc(self, doc_name: str) -> dict:
        url = f"{self.api_address}/swagger/{doc_name}/swagger.json"
        response = self._session.get(url=url)
        GatewayUtils.handle_response(response=response, return_codes=[200])
        return response.json()

    @staticmethod
    def _get_headers(access_token: str) -> [Dict]:
        return {"Authorization": "Bearer {}".format(access_token)}



"""
Container for Status classes used in cs18-api-client.
"""


class AppInstanceDeploymentStatus:
    PENDING = "Pending"
    DEPLOYING = "Deploying"
    CONFIGURING = "Configuring"
    DONE = "Done"
    ERROR = "Error"
    ABORTED = "Aborted"


class AppDeploymentStatus:
    PENDING = "Pending"
    DEPLOYING = "Deploying"
    DONE = "Done"
    ERROR = "Error"
    ABORTED = "Aborted"


class ServiceStatus:
    PENDING = "Pending"
    SETUP = "Setup"
    DONE = "Done"
    SETUP_FAILED = "SetupFailed"
    ABORTED = "Aborted"
    TERMINATING = "Terminating"
    TERMINATE_FAILED = "TerminateFailed"
    TERMINATED = "Terminated"
    VALIDATION_FAILED = "ValidationFailed"


class SandboxDeploymentStatus:
    WAITING = "Waiting"
    INITIATING = "Initiating"
    DEPLOYING = "Deploying"
    DONE = "Done"
    ERROR = "LaunchFailed"
    ENDING = "EndingSandbox"
    ENDING_FAILED = "EndingFailed"
    ABORTED = "ManuallyAborted"


class SandboxStatus:
    LAUNCHING = "Launching"
    ACTIVE = "Active"
    ACTIVE_WITH_ERROR = "ActiveWithError"
    ENDING = "Ending"
    ENDING_FAILED = "EndingFailed"
    ENDED = "Ended"
    ENDED_WITH_ERROR = "EndedWithError"
    NOT_FOUND = "NotFound"


class LogOutputTypes:
    INITIALIZATION = "initialization"
    HEALTHCHECK = "healthcheck"
    EVENTS = "events"
    ALL = [INITIALIZATION, HEALTHCHECK, EVENTS]


class Achievements:
    cloud_account_added = "cloud_account_added"
    first_blueprint_published = "first_blueprint_published"
    first_sandbox_launched = "first_sandbox_launched"
    first_automation_sandbox = "first_automation_sandbox"


class ProductionUpdateStatus:
    NONE = "None"
    DEPLOYING_GREEN = "DeployingGreen"
    REMOVING_GREEN = "RemovingGreen"
    EXPOSING_GREEN = 'ExposingGreen'
    TRANSITIONING_TO_GREEN = "TransitioningToGreen"
    CLEANING_UP = "CleaningUp"


class ProductionUpdateStrategy:
    BLUE_GREEN = "full environment blue/green"


class UserInviteReasons:
    ADMIN_JOIN_ACCOUNT = "AdminJoinAccount"
    TEAM_MEMBER_JOIN_SPACE = "TeamMemberJoinSpace"
    ADMIN_AUTHENTICATE_CLOUD_ACCOUNT = "AdminAuthenticateCloudAccount"


class DebuggingServiceStatus:
    ON = 'on'
    OFF = 'off'
    TURNING_ON = 'turning_on'
    TURNING_OFF = 'turning_off'
    DISABLED = 'disabled'
    NOT_READY = 'not_ready'



import json
import os
import sys

from setuptools import setup


def get_build_info_path():
    root = os.path.dirname(__file__)
    return os.path.join(root, "build_info.json")


def get_build_info_json():
    build_info_path = get_build_info_path()
    with open(build_info_path) as build_info_file:
        build_info_str = build_info_file.read().strip()
        build_info__json = json.loads(build_info_str)
    return build_info__json


def get_version():
    build_info_json = get_build_info_json()
    version = build_info_json["version"]
    return version


def update_requires():
    requires = []
    
    with open("Pipfile.lock") as lock_file:
        lock_json = json.loads(lock_file.read())
    
    for item in lock_json["default"]:
        requires.append(item + lock_json["default"][item]["version"])
    
    return requires


if sys.version_info < (3, 6):
    sys.exit('Only Python 3.6 and above is suppoorted.')


setup(
    name="cs18-api-client",
    version=get_version(),
    author="Quali",
    author_email="support@qualisystems.com",
    install_requires=update_requires(),
    packages=["gateways"],
    include_package_data=True,
    classifiers=["Programming Language :: Python :: 3"],
)



from typing import List, Optional
from datetime import datetime
import dateutil.parser
from gateways.common.cs18_api_classes import AccessLink
from gateways.common.cs18_api_errors import SandboxError, Error
from gateways.common.cs18_api_requests import AwsCloudProviderSettingsModel
from gateways.common.cs18_api_responses import (
    SpaceAdmin,
    SpaceCloudAccount,
    GetSpacesResponse,
    CatalogForGetResponse,
    CatalogForGetAllResponse,
    BlueprintResponse,
    CloudAccountResponse,
    BlueprintInputResponse,
    SandboxResponse,
    ProductionBlueResponse, ProductionGreenResponse,
    SandboxOwnerResponse,
    SandboxResponseLean,
    ProductionResponseLean,
    CloudResponse,
    ApplicationResponse,
    ApplicationResponseLean,
    AppInstanceResponse,
    HyperlinkResponse,
    BlueprintApplicationResponse,
    BlueprintRegionResponse,
    BlueprintCloudResponse,
    TokenResponse,
    CatalogApplicationResponse,
    UserPermittedToSpaceResponse,
    SourceControlDetailsResponse,
    BlueprintFileResponse,
    UserInvitationResponse,
    RepositoryResponse,
    InputResponse,
    RoleListItemResponse,
    AccountStatusResponse,
    ArtifactRepoCreatedUser,
    AccountStatusUserDetails,
    ArtifactRepoResponse,
    UserForAllUsersResponse,
    ServiceResponse,
    BlueprintServiceResponse,
    LaunchingProgress, LaunchingProgressStep, CloudAccountInSpaceResponse, CloudAccountConnectivityResponse,
    CloudAccountInfraSetResponse, CloudAccountSubnetsResponse, RegionResponse, GetImagesResponse,
    GetInstanceTypesResponse, VerifyCloudProviderResponse, VerifyCloudProviderResourceResponse,
    AccountRepositoryResponse, BlueprintValidationResponse, BlueprintSourceResponse, CloudAccountCostResponse,
    ComputeServiceInAccountResponse, ComputeServiceInSpaceResponse, ComputeServiceSpaceSpecResponse,
    K8SComputeServiceSpaceInfraSettingsResponse, BlueprintComputeServiceResponse,
    AccountRepositoryDetailsBranchResponse, AccountRepositoryDetailsResponse,
    SpaceRepositoryResponse, DirectAwsSshResponse, DirectAzureSshResponse, DirectRdpResponse, ParameterStoreItem,
    ParameterStoreValue, ParameterStoreOrigin, LiteralValue, AwsSsmValue, CloudRegionResponse,
    CloudComputeServiceResponse, InvitedUserDetailsResponse, RepositoryProviderResponse, RepositoryRelationshipResponse,
    AccountExtraDetails, AccountOnboardingLimitsResponse)


class Converters:
    """
    Converts provided dictionaries into requested Response objects
    """

    @staticmethod
    def create_spaces_response(space: dict):
        admins = [
            SpaceAdmin(
                first_name=admin["first_name"],
                last_name=admin["last_name"],
                account_role=admin["account_role"],
            )
            for admin in space["admins"]
        ]
        space_cloud_accounts = [
            SpaceCloudAccount(
                name=cloud_account["name"], cloud_account_type=cloud_account["type"]
            )
            for cloud_account in space["cloud_accounts"]
        ]
        return GetSpacesResponse(
            name=space["name"],
            admins=admins,
            users_count=space["users_count"],
            cloud_accounts=space_cloud_accounts,
        )

    @staticmethod
    def create_catalog_blueprint_full_response(catalog: dict) -> CatalogForGetResponse:
        """
        Returns CatalogForGetResponse based on provided dictionary
        :param catalog: dict
        :return: CatalogForGetResponse
        """
        return CatalogForGetResponse(
            name=catalog["blueprint_name"],
            description=catalog["description"],
            url=catalog["url"],
            last_modified=catalog["last_modified"],
            modified_by=catalog["modified_by"],
            clouds=catalog["clouds"] if catalog["clouds"] else [],
            applications=[
                Converters._create_catalog_app_response(app)
                for app in catalog["applications"]
            ]
            if catalog["applications"]
            else [],
            links=[Converters._create_link_response(link) for link in catalog["links"]]
            if catalog["links"]
            else [],
            artifacts=catalog.get("artifacts", {}),
            errors=catalog.get("errors"),
        )

    @staticmethod
    def create_catalog_blueprint_response(catalog: dict) -> CatalogForGetAllResponse:
        """
        Returns CatalogForGetAllResponse based on provided dictionary
        :param catalog: dict
        :return: CatalogForGetAllResponse
        """
        return CatalogForGetAllResponse(
            name=catalog["blueprint_name"],
            description=catalog["description"],
            url=catalog["url"],
            last_modified=catalog["last_modified"],
            modified_by=catalog["modified_by"],
            clouds=catalog["clouds"] if catalog["clouds"] else [],
            links=[Converters._create_link_response(link) for link in catalog["links"]]
            if catalog["links"]
            else [],
            artifacts=catalog.get("artifacts", {}),
            errors=catalog.get("errors"),
        )

    @staticmethod
    def create_artifact_repo_response(obj: dict) -> ArtifactRepoResponse:
        created_by_json = obj["created_by"]
        return ArtifactRepoResponse(
            obj["type"],
            dateutil.parser.parse(obj["created_date"]),
            obj["location"],
            ArtifactRepoCreatedUser(
                created_by_json["email"],
                created_by_json["first_name"],
                created_by_json["last_name"],
            ),
        )

    @staticmethod
    def create_blueprint_validation_response(blueprint: dict) -> BlueprintValidationResponse:
        return BlueprintValidationResponse(
            name=blueprint["blueprint_name"],
            description=blueprint["description"],
            url=blueprint["url"],
            source=Converters.create_blueprint_validation_source_response(blueprint.get("source")),
            artifacts=blueprint.get("artifacts", {}),
            inputs=[Converters._create_blueprint_input_response(i) for i in blueprint.get("inputs") or []],
            errors=[Error(e) for e in blueprint["errors"]],
            services=[Converters._create_blueprint_service_response(x) for x in blueprint.get("services") or []],
            applications=[Converters._create_blueprint_app_response(x) for x in blueprint.get("applications") or []],
            clouds=[Converters._create_blueprint_cloud_response(cloud) for cloud in blueprint.get("clouds") or []]
        )

    @staticmethod
    def create_blueprint_validation_source_response(source: dict) -> Optional[BlueprintSourceResponse]:
        if not source:
            return None
        return BlueprintSourceResponse(
            branch=source.get("branch"),
            commit=source.get("commit")
        )

    @staticmethod
    def create_blueprint_response(blueprint: dict) -> BlueprintResponse:
        """
        Returns BlueprintResponse based on provided dictionary
        :param blueprint: dict
        :return: BlueprintResponse
        """
        return BlueprintResponse(
            name=blueprint["blueprint_name"],
            description=blueprint["description"],
            url=blueprint["url"],
            enabled=blueprint["enabled"],
            last_modified=blueprint["last_modified"],
            is_sample=blueprint["is_sample"],
            modified_by=blueprint["modified_by"],
            clouds=[
                Converters._create_blueprint_cloud_response(cloud)
                for cloud in blueprint["clouds"]
            ]
            if blueprint["clouds"]
            else [],
            applications=[
                Converters._create_blueprint_app_response(app)
                for app in blueprint["applications"]
            ]
            if blueprint["applications"]
            else [],
            links=[
                Converters._create_link_response(link) for link in blueprint["links"]
            ],
            artifacts=blueprint.get("artifacts", {}),
            inputs=[
                Converters._create_blueprint_input_response(i)
                for i in blueprint.get("inputs", [])
            ],
            errors=[Error(e) for e in blueprint["errors"]],
            services=[
                Converters._create_blueprint_service_response(app)
                for app in blueprint["services"]
            ]
            if blueprint["services"]
            else []
        )

    @staticmethod
    def _create_cloud_account_cost_response(cloud_account: dict) -> CloudAccountCostResponse:
        cost: dict = cloud_account.get('cost')
        if not cost:
            return CloudAccountCostResponse(None)

        return CloudAccountCostResponse(
            Converters._parse_datetime(cost.get("last_update")))

    @staticmethod
    def _create_errors(cloud_account: dict) -> List[Error]:
        errors: List[dict] = cloud_account.get('errors', [])
        return [Error(e) for e in errors]

    @staticmethod
    def _create_compute_services_in_account_response(
            cloud_compute_services: List[dict]
    ) -> List[ComputeServiceInAccountResponse]:
        compute_services = [
            ComputeServiceInAccountResponse(
                name=compute_service.get("name", None),
                service_type=compute_service.get("type", None),
                created_date=Converters._parse_datetime(compute_service.get("created_date", None)),
                created_by=compute_service.get("created_by", None),
                user_defined=compute_service.get("user_defined", None),
                status=compute_service.get("status", None),
                spaces=compute_service.get("spaces", [])
            )
            for compute_service in cloud_compute_services
        ]
        return compute_services

    @staticmethod
    def _create_compute_services_in_space_response(
            cloud_compute_services: List[dict]
    ) -> List[ComputeServiceInSpaceResponse]:
        compute_services = [
            ComputeServiceInSpaceResponse(
                name=compute_service.get("name", None),
                service_type=compute_service.get("type", None),
                created_date=Converters._parse_datetime(compute_service.get("created_date", None)),
                created_by=compute_service.get("created_by", None),
                user_defined=compute_service.get("user_defined", None),
                status=compute_service.get("status", None),
                spec=Converters._create_compute_service_space_spec_response(compute_service.get("spec"))
            )
            for compute_service in cloud_compute_services
        ]
        return compute_services

    @staticmethod
    def _create_compute_service_space_spec_response(spec: dict) -> Optional[ComputeServiceSpaceSpecResponse]:
        if spec is None:
            return None
        return ComputeServiceSpaceSpecResponse(
            kubernetes=Converters._create_k8s_compute_service_space_infra_settings_response(spec.get("kubernetes"))
        )

    @staticmethod
    def _create_k8s_compute_service_space_infra_settings_response(kubernetes: dict) \
            -> Optional[K8SComputeServiceSpaceInfraSettingsResponse]:
        if kubernetes is None:
            return None
        return K8SComputeServiceSpaceInfraSettingsResponse(
            namespace=kubernetes.get("namespace"),
            internet_facing=kubernetes.get("internet_facing")
        )

    @staticmethod
    def create_cloud_account_response(cloud: dict) -> CloudAccountResponse:
        """
        Returns CloudAccountResponse based on provided dictionary
        :param cloud: dict
        :return: CloudAccountResponse
        """
        return CloudAccountResponse(
            name=cloud["name"],
            provider=cloud["type"],
            created_date=Converters._parse_datetime(cloud.get("created_date")),
            created_by=cloud["created_by"],
            spaces=cloud["spaces"],
            compute_services=Converters._create_compute_services_in_account_response(
                cloud.get("compute_services", None)
            ),
            cost=Converters._create_cloud_account_cost_response(cloud),
            errors=Converters._create_errors(cloud)
        )

    @staticmethod
    def create_cloud_account_in_space_response(cloud: dict) -> CloudAccountInSpaceResponse:
        """
        Returns CloudAccountResponse based on provided dictionary
        :param cloud: dict
        :return: CloudAccountResponse
        """
        return CloudAccountInSpaceResponse(
            name=cloud["name"],
            provider=cloud["type"],
            created_date=Converters._parse_datetime(cloud.get("created_date")),
            created_by=cloud["created_by"],
            internet_facing=cloud["internet_facing"],
            connectivity=Converters._create_cloud_account_connectivity_response(cloud.get("connectivity")),
            compute_services=Converters._create_compute_services_in_space_response(
                cloud.get("compute_services", None)
            ),
            cost=Converters._create_cloud_account_cost_response(cloud),
            errors=Converters._create_errors(cloud)
        )

    @staticmethod
    def _create_cloud_account_connectivity_response(conn: dict) -> Optional[CloudAccountConnectivityResponse]:
        if not conn:
            return None

        return CloudAccountConnectivityResponse(
            existing_infra=conn.get("existing_infra"),
            existing_infra_sets=[CloudAccountInfraSetResponse(
                region=infra_set["region"],
                virtual_network=infra_set["virtual_network"],
                subnets=CloudAccountSubnetsResponse(
                    gateway_subnet=infra_set["subnets"].get("gateway_subnet"),
                    management_subnet=infra_set["subnets"].get("management_subnet"),
                    application_subnets=infra_set["subnets"].get("application_subnets")
                ) if "subnets" in infra_set else None) for infra_set in conn.get("existing_infra_sets")]
            if "existing_infra_sets" in conn else None)

    @staticmethod
    def _create_blueprint_input_response(
            blueprint_input: dict
    ) -> BlueprintInputResponse:
        return BlueprintInputResponse(
            name=blueprint_input["name"],
            default_value=blueprint_input.get("default_value"),
            display_style=blueprint_input.get("display_style"),
            optional=blueprint_input.get("optional", False),
            description=blueprint_input.get("description"),
        )

    @staticmethod
    def create_sandbox_response(sandbox: dict) -> SandboxResponse:
        """
        Returns SandboxResponse based on provided dictionary
        :param sandbox: dict
        :return: SandboxResponse
        """
        return SandboxResponse(
            sandbox_id=sandbox["id"],
            blueprint=sandbox["blueprint_name"],
            applications=[
                Converters._create_app_response(service)
                for service in sandbox["applications"]
            ]
            if sandbox["applications"]
            else [],
            services=[
                Converters._create_service_response(s)
                for s in sandbox.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(sandbox.get("inputs")),
            artifacts=sandbox.get("artifacts"),
            automation=sandbox["automation"],
            start_time=sandbox["deployment_start_time"],
            create_time=sandbox["create_time"],
            status_details=sandbox["status_details"],
            sandbox_status=sandbox["sandbox_status"],
            status_error_description=sandbox["status_error_description"],
            links=[Converters._create_link_response(link) for link in sandbox["links"]]
            if sandbox["links"]
            else [],
            owner=Converters._create_sandbox_owner(sandbox["owner"])
            if sandbox["owner"]
            else [],
            errors=[
                Converters._create_sandbox_error(error)
                for error in sandbox.get("errors", [])
            ],
            scheduled_end_time=dateutil.parser.parse(sandbox["scheduled_end_time"])
            if sandbox.get("scheduled_end_time", None) else None,
            end_time=dateutil.parser.parse(sandbox["end_time"])
            if sandbox.get("end_time", None) else None,
            debugging_service=sandbox.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(sandbox.get('launching_progress')),
            source=Converters._create_source(sandbox.get('source')),
            direct_access=sandbox.get('direct_access', 'off')
        )

    @staticmethod
    def create_production_blue_response(production: dict) -> ProductionBlueResponse:
        """
        Returns ProductionBlueResponse based on provided dictionary
        :param production: dict
        :return: ProductionBlueResponse
        """
        return ProductionBlueResponse(
            production_id=production["id"],
            sandbox_id=production["sandbox_id"],
            name=production["name"],
            owner=Converters._create_sandbox_owner(production["owner"]),
            blueprint_name=production["blueprint_name"],
            blueprint_description=production["blueprint_description"],
            build=production["build"],
            applications=[
                Converters._create_app_response(app)
                for app in production.get("applications", [])
            ],
            services=[
                Converters._create_service_response(s)
                for s in production.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(production.get("inputs")),
            artifacts=production.get("artifacts"),
            deployment_start_time=Converters._parse_datetime(
                production.get("deployment_start_time")
            ),
            create_time=Converters._parse_datetime(production.get("create_time")),
            status_details=production["status_details"],
            production_status=production["production_status"],
            update_status=production["update_status"],
            links=[
                Converters._create_link_response(link)
                for link in production.get("links") or []
            ],
            errors=[
                Converters._create_sandbox_error(error)
                for error in production.get("errors") or []
            ],
            status_error_description=production["status_error_description"],
            has_green=production["has_green"],
            debugging_service=production.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(production.get('launching_progress')),
            direct_access=production.get('direct_access', 'off')
        )

    @staticmethod
    def create_production_green_response(production: dict) -> ProductionGreenResponse:
        """
        Returns ProductionGreenResponse based on provided dictionary
        :param production: dict
        :return: ProductionGreenResponse
        """
        return ProductionGreenResponse(
            production_id=production["id"],
            sandbox_id=production["sandbox_id"],
            name=production["name"],
            owner=Converters._create_sandbox_owner(production["owner"]),
            blueprint_name=production["blueprint_name"],
            blueprint_description=production["blueprint_description"],
            build=production["build"],
            applications=[Converters._create_app_response(app) for app in production.get("applications", [])],
            services=[
                Converters._create_service_response(s)
                for s in production.get("services", [])
            ],
            inputs=Converters._create_sandbox_inputs(production.get("inputs")),
            artifacts=production.get("artifacts"),
            deployment_start_time=Converters._parse_datetime(production.get("deployment_start_time")),
            create_time=Converters._parse_datetime(production.get("create_time")),
            status_details=production["status_details"],
            production_status=production["production_status"],
            update_status=production["update_status"],
            links=[Converters._create_link_response(link) for link in production.get("links") or []],
            errors=[Converters._create_sandbox_error(error) for error in production.get("errors") or []],
            status_error_description=production["status_error_description"],
            exposure=production["exposure"],
            debugging_service=production.get('debugging_service'),
            launching_progress=Converters._create_launching_progress(production.get('launching_progress')),
            direct_access=production.get('direct_access', 'off')
        )

    @staticmethod
    def _create_sandbox_owner(user: dict) -> Optional[SandboxOwnerResponse]:
        if not user:
            return None

        return SandboxOwnerResponse(
            first_name=user["first_name"],
            last_name=user["last_name"],
            email=user["email"],
            join_date=user["join_date"],
        )

    @staticmethod
    def _create_sandbox_error(sandbox_error: dict) -> SandboxError:
        return SandboxError(
            time=sandbox_error["time"],
            code=sandbox_error["code"],
            message=sandbox_error["message"],
        )

    @staticmethod
    def _create_source(source: dict) -> Optional[BlueprintSourceResponse]:
        if not source:
            return None
        return BlueprintSourceResponse(branch=source.get('branch'), commit=source.get('commit'))

    @staticmethod
    def create_sandbox_response_list_item(sandbox: dict) -> SandboxResponseLean:
        """
        Returns SandboxResponseLean based on provided dictionary
        :param sandbox: dict
        :return: SandboxResponseLean
        """
        return SandboxResponseLean(
            sandbox_id=sandbox["id"],
            blueprint=sandbox["blueprint_name"],
            applications=[
                Converters._create_application_response_lean(application)
                for application in sandbox["applications"]
            ]
            if sandbox["applications"]
            else [],
            services=[
                Converters._create_service_response(s)
                for s in sandbox.get("services", [])
            ],
            automation=sandbox["automation"],
            start_time=sandbox["deployment_start_time"],
            create_time=sandbox["create_time"],
            status_details=sandbox["status_details"],
            sandbox_status=sandbox["sandbox_status"],
            status_error_description=sandbox["status_error_description"],
            links=[Converters._create_link_response(link) for link in sandbox["links"]]
            if sandbox["links"]
            else [],
            owner=Converters._create_sandbox_owner(sandbox["owner"])
            if sandbox["owner"]
            else [],
            name=sandbox["name"],
            errors=[
                Converters._create_sandbox_error(error)
                for error in sandbox.get("errors", [])
            ],
            scheduled_end_time=dateutil.parser.parse(sandbox["scheduled_end_time"]) if sandbox.get(
                "scheduled_end_time") else None,
            end_time=dateutil.parser.parse(sandbox["end_time"]) if sandbox.get("end_time") else None,
            debugging_service=sandbox.get("debugging_service"),
            source=Converters._create_source(sandbox.get('source')),
            direct_access=sandbox.get('direct_access', 'off')
        )
 
    @staticmethod
    def create_production_response_list_item(
            production: dict
    ) -> ProductionResponseLean:
        """
        Returns ProductionResponseLean based on provided dictionary
        :param production: dict
        :return: ProductionResponseLean
        """
        return ProductionResponseLean(
            production_id=production["id"],
            name=production["name"],
            blueprint_description=production["blueprint_description"],
            blue_status=production["blue_status"],
            green_status=production["green_status"],
            update_strategy=production["update_strategy"],
            update_status=production["update_status"],
            create_time=Converters._parse_datetime(production.get("create_time")),
            deployment_start_time=Converters._parse_datetime(production.get("deployment_start_time")),
            blue_applications=[Converters._create_application_response_lean(x)
                               for x in production.get("blue_applications") or {}],
            blue_services=[Converters._create_service_response(s) for s in production.get("services", [])],
            direct_access=production.get('direct_access', 'off')
        )

    @staticmethod
    def _create_cloud_region_response(region: dict) -> CloudRegionResponse:
        if not region:
            return CloudRegionResponse("", "")
        return CloudRegionResponse(region_id=region.get("id"), region_name=region.get("name"))

    @staticmethod
    def _create_cloud_compute_service(compute_service: dict) -> CloudComputeServiceResponse:
        return CloudComputeServiceResponse(
            name=compute_service.get("name", None),
            service_type=compute_service.get("type", None),
            created_date=Converters._parse_datetime(compute_service.get("created_date", None)),
            created_by=compute_service.get("created_by", None),
            user_defined=compute_service.get("user_defined", None)
        )

    @staticmethod
    def _create_cloud_response(cloud: dict) -> CloudResponse:
        return CloudResponse(
            provider=cloud.get("provider"),
            cloud_account_name=cloud.get("cloud_account_name"),
            region=Converters._create_cloud_region_response(cloud.get("region")),
            compute_service=Converters._create_cloud_compute_service(cloud.get("compute_service"))
        )

    @staticmethod
    def _create_app_response(application: dict) -> ApplicationResponse:
        return ApplicationResponse(
            name=application["name"],
            shortcuts=application["shortcuts"],
            status=application["status"],
            instances=[
                Converters._create_instance_response(instance)
                for instance in application["instances"]
            ],
            private_address=application["private_address"],
            public_address=application["public_address"],
            internal_ports=application["internal_ports"],
            external_ports=application["external_ports"],
            internal_dns=application["internal_dns"],
            image_name=application["image_name"],
            cloud=Converters._create_cloud_response(application.get("cloud")),
            icon=application.get("icon")

            if application.get("cloud")
            else None,
            direct_access_protocols=application.get("direct_access_protocols", [])
        )

    @staticmethod
    def _create_application_response_lean(application: dict) -> ApplicationResponseLean:
        return ApplicationResponseLean(
            name=application["name"],
            status=application["status"],
            cloud=Converters._create_cloud_response(application.get("cloud"))
            if application.get("cloud")
            else None,
        )

    @staticmethod
    def _create_instance_response(instance: dict) -> AppInstanceResponse:
        access_links = None
        if "access_links" in instance and instance["access_links"] is not None:
            access_links = [
                Converters._create_access_link(al) for al in instance["access_links"]
            ]

        return AppInstanceResponse(
            private_ip=instance["private_ip"],
            infra_id=instance["infra_id"],
            instance_type=instance["instance_type"],
            memory_limit=instance["memory_limit"],
            cpu_limit=instance["cpu_limit"],
            deployment_status=instance["status"],
            access_links=access_links,
            links=[Converters._create_link_response(link) for link in instance["links"]]
            if instance["links"]
            else [],
        )

    @staticmethod
    def _create_access_link(access_link_json: dict) -> AccessLink:
        return AccessLink(
            protocol=access_link_json["protocol"], link=access_link_json["link"]
        )

    @staticmethod
    def _create_link_response(link: dict) -> HyperlinkResponse:
        return HyperlinkResponse(
            rel=link["rel"], href=link["href"], method=link["method"]
        )

    @staticmethod
    def _create_blueprint_app_response(app: dict) -> BlueprintApplicationResponse:

        return BlueprintApplicationResponse(name=app["name"], version=app["version"])

    @staticmethod
    def _create_blueprint_service_response(service: dict) -> BlueprintServiceResponse:

        return BlueprintServiceResponse(name=service["name"], service_type=service["type"])

    @staticmethod
    def _create_blueprint_cloud_region_response(
            region: dict
    ) -> BlueprintRegionResponse:
        if not region:
            return BlueprintRegionResponse("", "")
        return BlueprintRegionResponse(
            region_id=region.get("id", ""), name=region.get("name", "")
        )

    @staticmethod
    def _create_blueprint_cloud_response(cloud):
        if isinstance(cloud, str):
            return cloud
        return BlueprintCloudResponse(
            cloud_account_name=cloud.get("cloud_account_name"),
            provider=cloud.get("provider"),
            region=Converters._create_blueprint_cloud_region_response(
                cloud.get("region")
            ),
            compute_service=Converters._create_blueprint_compute_service(cloud.get("compute_service"))
        )

    @staticmethod
    def create_token_response(link: dict) -> TokenResponse:
        """
        Returns TokenResponse based on provided dictionary
        :param link: dict
        :return: TokenResponse
        """
        return TokenResponse(
            access_token=link["access_token"],
            refresh_token=link["refresh_token"],
            token_type=link["token_type"],
            expires_in=link["expires_in"],
        )

    @staticmethod
    def _create_catalog_app_response(app: dict) -> CatalogApplicationResponse:
        return CatalogApplicationResponse(name=app["name"], version=app["version"])

    @staticmethod
    def create_user_permitted_to_space_response(
            user: dict
    ) -> UserPermittedToSpaceResponse:
        """
        Returns UserPermittedToSpaceResponse based on provided dictionary
        :param user: dict
        :return: UserPermittedToSpaceResponse
        """
        return UserPermittedToSpaceResponse(
            first_name=user["first_name"],
            last_name=user["last_name"],
            email=user["email"],
            join_date=user["join_date"],
            space_role=user["space_role"],
            account_role=user["account_role"],
            has_access_to_all_spaces=user["has_access_to_all_spaces"]
        )

    @staticmethod
    def create_user_for_all_users_response(
            user: dict
    ) -> UserForAllUsersResponse:
        """
        Returns UserForAllUsersResponse based on provided dictionary
        :param user: dict
        :return: UserForAllUsersResponse
        """
        return UserForAllUsersResponse(
            email=user["email"],
            first_name=user["first_name"],
            last_name=user["last_name"],
            join_date=user["join_date"],
            account_role=user["account_role"],
            has_access_to_all_spaces=user["has_access_to_all_spaces"],
            member_in_spaces=user["member_in_spaces"]
        )

    @staticmethod
    def _create_source_control_details_response(
            source_control: dict
    ) -> SourceControlDetailsResponse:
        return SourceControlDetailsResponse(
            branch_name=source_control["branch_name"],
            latest_commit_id=source_control["latest_commit_id"],
            changed_by=source_control["changed_by"],
            changed_date=dateutil.parser.parse(source_control["change_date"]),
        )

    @staticmethod
    def create_blueprint_file_response(file: dict) -> BlueprintFileResponse:
        """
        Returns BlueprintFileResponse based on provided dictionary
        :param file: dict
        :return: BlueprintFileResponse
        """
        return BlueprintFileResponse(
            url=file["url"],
            kind=file["kind"],
            content=file["content"],
            source_control_details=Converters._create_source_control_details_response(
                file["source_control_details"]
            ),
            icon=file.get("icon"),
        )

    @staticmethod
    def create_user_invitation_response(invite: dict) -> UserInvitationResponse:
        """
        Returns UserInvitationResponse based on provided dictionary
        :param invite: dict
        :return: UserInvitationResponse
        """
        return UserInvitationResponse(
            email=invite["email"],
            send_date=invite["send_date"],
            account_role=invite["account_role"],
            space_name=invite["space_name"],
            space_role=invite["space_role"],
        )

    @staticmethod
    def create_invited_user_details_response(invited_user_details: dict) -> InvitedUserDetailsResponse:
        """
        Returns InvitedUserDetailsResponse based on provided dictionary
        :param invited_user_details: dict
        :return: InvitedUserDetailsResponse
        """
        return InvitedUserDetailsResponse(
            email=invited_user_details["email"],
            account_name=invited_user_details["account_name"],
        )

    @staticmethod
    def create_repository_response(repo: dict) -> RepositoryResponse:
        """
        Returns RepositoryResponse based on provided dictionary
        :param repo: dict
        :return: RepositoryResponse
        """
        return RepositoryResponse(
            repository_url=repo["repository_url"],
            errors=[Error(e) for e in repo["errors"]],
            branch=repo['branch']
        )

    @staticmethod
    def create_repository_provider_response(repo: dict) -> RepositoryProviderResponse:
        """
        Returns RepositoryProviderResponse based on provided dictionary
        :param repo: dict
        :return: RepositoryProviderResponse
        """

        return RepositoryProviderResponse(
            provider_id=repo["id"],
            name=repo["name"],
            details=repo["details"],
            status=repo["status"],
            account_id=repo["account_id"],
            repositories=[Converters.create_account_repository_response for r in repo["repositories"]],
            errors=[Error(e) for e in repo["errors"]],
        )

    @staticmethod
    def create_account_repository_response(repo: dict) -> AccountRepositoryResponse:
        """
        Returns AccountRepositoryResponse based on provided dictionary
        :param repo: dict
        :return: AccountRepositoryResponse
        """
        return AccountRepositoryResponse(
            repository_id=repo["id"],
            name=repo["name"],
            provider_id=repo["provider_id"],
            allow_sharing=repo["allow_sharing"],
            open_access=repo["open_access"],
            account_id=repo["account_id"],
            is_valid=repo["is_valid"],
            repository_url=repo["repository_url"],
            connection_details=repo["connection_details"],
            errors=[Error(e) for e in repo["errors"]],
        )

    @staticmethod
    def create_repository_relationship_response(repo: dict) -> RepositoryRelationshipResponse:
        """
        Returns RepositoryRelationshipResponse based on provided dictionary
        :param repo: dict
        :return: RepositoryRelationshipResponse
        """

        return RepositoryRelationshipResponse(
            repo_relationship_id=repo["id"],
            space_id=repo["space_id"],
            deployment_type=repo["deployment_type"],
            branch=repo["branch"],
            account_id=repo["account_id"],
            is_valid=repo["is_valid"],
            errors=[Error(e) for e in repo["errors"]],
        )

    @staticmethod
    def create_role_list_item_response(role: dict) -> RoleListItemResponse:
        return RoleListItemResponse(name=role["name"])

    @staticmethod
    def _create_sandbox_inputs(response_inputs: dict) -> List[InputResponse]:
        inputs = []
        if response_inputs:
            for request_input in response_inputs:
                inputs.append(
                    InputResponse(
                        name=request_input["name"],
                        value=request_input["value"],
                        description=request_input["description"],
                        display_style=request_input["display_style"],
                        optional=request_input.get("optional", False),
                    )
                )
        return inputs

    @staticmethod
    def create_account_status_response(obj: dict):
        return AccountStatusResponse(
            AccountOnboardingLimitsResponse(obj["limits"]),
            obj["onboarding_status"],
            Converters._parse_datetime(obj["license_expiration_date"]),
            obj["ignore_expiration_date"],
            obj["suspended_reason"],
            obj["salesforce_account_id"],
            obj["plan_id"],
            AccountStatusUserDetails(obj["user_details"]),
        )

    @staticmethod
    def create_account_extra_details(json: dict):
        return AccountExtraDetails(
            account_id=json['id'],
            name=json['name'],
            sso_enabled=json['sso_enabled'],
            anonymous_users=json['anonymous_users'],
            sso_update_enabled=json['sso_update_enabled']
        )

    @staticmethod
    def _parse_datetime(value: str) -> datetime:
        if value:
            return dateutil.parser.parse(value)
        return None

    @staticmethod
    def _create_service_response(service: dict) -> ServiceResponse:
        return ServiceResponse(
            name=service["name"],
            status=service["status"],
            service_type=service["type"]
        )

    @staticmethod
    def _create_launching_progress(progress: dict) -> Optional[LaunchingProgress]:
        if not progress:
            return None
        return LaunchingProgress(
            creating_infrastructure=Converters._create_launching_progress_step(progress.get('creating_infrastructure')),
            preparing_artifacts=Converters._create_launching_progress_step(progress.get('preparing_artifacts')),
            deploying_applications=Converters._create_launching_progress_step(progress.get('deploying_applications')),
            verifying_environment=Converters._create_launching_progress_step(progress.get('verifying_environment')))

    @staticmethod
    def _create_launching_progress_step(step: dict) -> LaunchingProgressStep:
        return LaunchingProgressStep(
            total=step['total'],
            succeeded=step['succeeded'],
            failed=step['failed'],
            status=step['status'])

    @staticmethod
    def create_space_cloud_account_regions_response(region: dict) -> RegionResponse:
        return RegionResponse(region["id"], region["name"])

    @staticmethod
    def get_image_response(images: dict):
        return GetImagesResponse(images['image_id'], images['image_name'])

    @staticmethod
    def get_instance_type_response(instance_types: dict):
        return GetInstanceTypesResponse(instance_types['instance_type_name'])

    @staticmethod
    def _create_resources_summary_response(resources_summary: dict) -> List[VerifyCloudProviderResourceResponse]:
        resources = []
        if resources_summary:
            for resource in resources_summary:
                resources.append(
                    VerifyCloudProviderResourceResponse(
                        resource_type=resource["resource_type"],
                        resource_name=resource["resource_name"],
                        resource_id=resource["resource_id"],
                        resource_status=resource["resource_status"],
                        resource_valid=resource["resource_valid"]
                    )
                )
        return resources

    @staticmethod
    def create_verify_response(verify: dict) -> VerifyCloudProviderResponse:
        return VerifyCloudProviderResponse(
            success=verify["success"],
            resources_summary=Converters._create_resources_summary_response(verify["resources_summary"]),
        )

    @staticmethod
    def create_account_repository_details_branch_response(data: dict) -> AccountRepositoryDetailsBranchResponse:
        return AccountRepositoryDetailsBranchResponse(name=data['name'])

    @staticmethod
    def create_account_repository_details_response(data: dict) -> AccountRepositoryDetailsResponse:
        return AccountRepositoryDetailsResponse(repository_name=data['name'],
                                                repository_full_name=data['full_name'],
                                                default_branch=data['default_branch'],
                                                branches=[AccountRepositoryDetailsBranchResponse(b)
                                                          for b in data["branches"]])

    @staticmethod
    def _create_blueprint_compute_service(compute_service: dict) -> BlueprintComputeServiceResponse:
        return BlueprintComputeServiceResponse(
            name=compute_service.get("name", None),
            service_type=compute_service.get("type", None),
            created_date=Converters._parse_datetime(compute_service.get("created_date", None)),
            created_by=compute_service.get("created_by", None),
            user_defined=compute_service.get("user_defined", None)
        )

    @staticmethod
    def create_space_repository_response(data: dict) -> SpaceRepositoryResponse:
        return SpaceRepositoryResponse(repository_name=data['name'],
                                       repository_type=data['repository_type'],
                                       repository_url=data['repository_url'],
                                       branch=data['branch'])

    @staticmethod
    def create_direct_aws_ssh_response(data: dict) -> DirectAwsSshResponse:
        return DirectAwsSshResponse(
            connection=data["connection"],
            pem_file_url=data["pem_file_url"])

    @staticmethod
    def create_direct_azure_ssh_response(data: dict) -> DirectAzureSshResponse:
        return DirectAzureSshResponse(
            connection=data["connection"],
            password=data["password"])

    @staticmethod
    def create_direct_rdp_response(data: dict) -> DirectRdpResponse:
        return DirectRdpResponse(
            rdp_file_url=data["rdp_file_url"],
            username=data["username"],
            password=data["password"])

    @staticmethod
    def create_parameter_value_response(origin: str, raw_parameter_value: dict) -> ParameterStoreValue:
        if origin == ParameterStoreOrigin.Literal:
            return LiteralValue(content=raw_parameter_value["content"])
        if origin == ParameterStoreOrigin.AwsSsm:
            return AwsSsmValue(path=raw_parameter_value["path"])

        return None

    @staticmethod
    def create_get_all_parameters_response(data: dict) -> List[ParameterStoreItem]:
        parameters = []
        if data:
            for parameter in data:
                parameters.append(
                    ParameterStoreItem(
                        name=parameter["name"],
                        origin=parameter["origin"],
                        value=Converters.create_parameter_value_response(parameter["origin"], parameter["value"]),
                        description=parameter["description"],
                        modified_by=parameter["modified_by"],
                        last_modified=parameter["last_modified"])
                )
        return parameters

    @staticmethod
    def create_get_parameter_response(data: dict) -> Optional[ParameterStoreItem]:
        parameter = None
        if data:
            parameter = ParameterStoreItem(
                name=data["name"],
                origin=data["origin"],
                value=Converters.create_parameter_value_response(data["origin"], data["value"]),
                description=data["description"],
                modified_by=data["modified_by"],
                last_modified=data["last_modified"])
        return parameter

    @staticmethod
    def aws_cloud_provider_settings_model(data: dict) -> AwsCloudProviderSettingsModel:
        return AwsCloudProviderSettingsModel(
            sidecar_image_param_name=data.get("sidecar_image_param_name", None)
        )
